use std::error::Error;

use futures::{SinkExt, StreamExt};
use tokio_tungstenite::tungstenite::Message;

use crate::data::{BoatsLayout, GameRules};
use crate::human_player_ws::{ClientMessage, ServerMessage};

#[derive(Debug, Eq, PartialEq)]
pub enum ClientEndResult {
    Finished,
    InvalidBoatsLayout,
}

pub async fn run_client(
    server: &str,
    requested_rules: &GameRules,
    layout: BoatsLayout,
) -> Result<ClientEndResult, Box<dyn Error>> {
    let url = format!(
        "{}/play/bot?{}",
        server.replace("http", "ws"),
        serde_urlencoded::to_string(requested_rules).unwrap()
    );
    log::debug!("Connecting to  {}...", url);
    let (mut socket, _) = match tokio_tungstenite::connect_async(url).await {
        Ok(s) => s,
        Err(e) => {
            log::error!("Failed to establish WebSocket connection! {:?}", e);
            return Err(Box::new(e));
        }
    };

    while let Some(chunk) = socket.next().await {
        let message = match chunk? {
            Message::Text(message) => {
                log::debug!("TEXT message from server: {}", message);

                let msg: ServerMessage = serde_json::from_str(&message)?;
                msg
            }
            Message::Binary(_) => {
                log::debug!("BINARY message from server");
                continue;
            }
            Message::Ping(_) => {
                log::debug!("PING from server");
                continue;
            }
            Message::Pong(_) => {
                log::debug!("PONG from server");
                continue;
            }
            Message::Close(_) => {
                log::debug!("CLOSE message request from server");
                break;
            }
            Message::Frame(_) => {
                log::debug!("Frame from server");
                continue;
            }
        };

        match message {
            ServerMessage::WaitingForAnotherPlayer => log::debug!("Waiting for other player..."),
            ServerMessage::QueryBoatsLayout { rules } => {
                assert_eq!(&rules, requested_rules);
                log::debug!("Server requested boats layout");
                socket
                    .send(Message::Text(serde_json::to_string(
                        &ClientMessage::BoatsLayout {
                            layout: layout.clone(),
                        },
                    )?))
                    .await?;
            }
            ServerMessage::WaitingForOtherPlayerConfiguration => {
                log::debug!("Waiting for other player configuration...")
            }
            ServerMessage::OtherPlayerReady => log::debug!("Other player is ready!"),
            ServerMessage::GameStarting => log::debug!("The game is starting..."),
            ServerMessage::OtherPlayerMustFire { .. } => log::debug!("Other player must fire!"),
            ServerMessage::RequestFire { status } => {
                let location = status.find_valid_random_fire_location();
                log::debug!("Will fire at {:?}", location);
                socket
                    .send(Message::Text(serde_json::to_string(
                        &ClientMessage::Fire { location },
                    )?))
                    .await?;
            }
            ServerMessage::StrikeResult { pos, result } => {
                log::debug!("Strike at {} result: {:?}", pos.human_print(), result)
            }
            ServerMessage::OpponentStrikeResult { pos, result } => log::debug!(
                "Opponent trike at {} result: {:?}",
                pos.human_print(),
                result
            ),
            ServerMessage::LostGame {
                your_map,
                opponent_map,
            } => {
                log::debug!("We lost game :(");
                log::debug!("Other game:\n{}\n", opponent_map.get_map());
                log::debug!("Our game:\n{}\n", your_map.get_map());
                break;
            }
            ServerMessage::WonGame {
                your_map,
                opponent_map,
            } => {
                log::debug!("We won the game !!!!");
                log::debug!("Other game:\n{}\n", opponent_map.get_map());
                log::debug!("Our game:\n{}\n", your_map.get_map());
                break;
            }
            ServerMessage::RejectedBoatsLayout { errors } => {
                log::warn!("Rejected boat layout: {:?}", errors);
                return Ok(ClientEndResult::InvalidBoatsLayout);
            }
        }
    }

    Ok(ClientEndResult::Finished)
}
